@@ -1,5 +1,6 @@
 # See LICENSE for license details.
 
+import pathlib
 import argparse
 import os
 import sys
@@ -7,9 +8,9 @@ import pkg_resources
 import tarfile
 import shutil
 import yaml
-import chisel3_jar
 import mill_bin
 import mill_cache
+import glob
 
 # Check if this is run from a local installation
 scratchipdir = os.path.abspath(
@@ -20,6 +21,10 @@ if os.path.exists(os.path.join(scratchipdir, "scratchip")):
 
 from scratchip import __version__
 
+def read_lines(path):
+    with open(path,'r') as f:
+        return f.readlines()
+
 def get_resource_name(name):
     return pkg_resources.resource_filename(__name__, name)
 
@@ -30,16 +35,17 @@ class ScratChip:
     prj_name = "None"
     prj_path = "None"
     top_name = "None"
-    scratchip_path = 'builds/scratchip'
+    scratchip_path = '.scratchip'
     cfg = None
 
     def __init__(self, prj_name, cfg):
         self.prj_name = prj_name
         self.prj_path = os.path.abspath(self.prj_name)
         with open(cfg) as file:
-            if sys.version_info.major == 3 and sys.version_info.minor > 6:
+            if sys.version_info.major == 3 and sys.version_info.minor >= 6:
                 self.cfg = yaml.load(file, Loader=yaml.FullLoader)
             else:
+                print(sys.version_info)
                 self.cfg = yaml.load(file)
 
     def create(self, top):
@@ -55,7 +61,6 @@ class ScratChip:
         cache_path = os.path.join(self.prj_path, self.scratchip_path)
         mill_path = os.path.join(self.prj_path, self.scratchip_path + '/mill')
         jars_path = os.path.join(self.prj_path, self.scratchip_path + '/jars')
-        chisel3_jar_path = os.path.join(jars_path, 'chisel3.jar')
 
         if not os.path.exists(cache_path):
             self.extract_cache(mill_cache.source, cache_path)
@@ -63,11 +68,15 @@ class ScratChip:
             shutil.copyfile(mill_bin.source, mill_path)
         if not os.path.exists(jars_path):
             os.makedirs(jars_path)
-        if not os.path.exists(chisel3_jar_path):
-            shutil.copyfile(chisel3_jar.source, chisel3_jar_path)
+
+        # create project's YAML template
+        prj_yml_dest = os.path.join(self.prj_path, 'project.yml')
+        if not os.path.exists(prj_yml_dest):
+            prj_yml = get_resource_name("assets/project.yml")
+            shutil.copyfile(prj_yml, prj_yml_dest)
 
     def dump_default_cfg(self, cfg, dump_name):
-       shutil.copyfile(args.config, args.dump_name)
+       shutil.copyfile(cfg, dump_name)
 
     def create_dir(self, path, dir_tree):
         for (k, v) in dir_tree.items():
@@ -138,28 +147,57 @@ class ScratChip:
 
     def read_yaml(self, cfg_path):
         with open(cfg_path) as file:
-            if sys.version_info.major == 3 and sys.version_info.minor > 6:
+            if sys.version_info.major == 3 and sys.version_info.minor >= 6:
                 return yaml.load(file, Loader=yaml.FullLoader)
             else:
                 return yaml.load(file)
 
     def gen_filelist_str(self, path_list):
-        res = []
+        inc_dir = []
+        res = {
+            "filelist" : [],
+            "dirs"     : [],
+            "files"    : [],
+        }
         for path in path_list:
             if isinstance(path, dict):
                 k, v = list(path.items())[0]
                 tag = ''
-                if v == 'is_include_file':
-                    tag = '-v'
+                if v == "flat_dir":
+                    flat_files = glob.glob(os.path.abspath(k) + "/*.*v")
+                    if not flat_files:
+                        print("Warn: %s is empty" % k)
+                    res["filelist"].extend([x + "\n" for x in flat_files])
+                    res["files"].extend([x + "\n" for x in flat_files])
+                elif v == 'flat_filelist':
+                    flat_files = read_lines(os.path.abspath(k))
+                    res["filelist"].extend(flat_files)
+                    res["files"].extend(flat_files)
                 elif v == 'is_include_dir':
-                    tag = '-y'
+                    tag = '+incdir+'
+                    inc_dir.append("%s%s\n" % (tag, os.path.abspath(k)))
+                    res["dirs"].append("%s" % os.path.abspath(k))
                 else:
-                    print("Unsupport tag: %s" % v)
-                    sys.exit(-1)
-                res.append("%s %s\n" % (tag, os.path.abspath(k)))
+                    if v == 'is_library_file':
+                        tag = '-v '
+                        res["files"].append("%s" % os.path.abspath(k))
+                    elif v == 'is_filelist':
+                        tag = '-f '
+                        res["files"].append("%s" % os.path.abspath(k))
+                    elif v == 'is_library_dir':
+                        tag = '-y '
+                        res["dirs"].append("%s" % os.path.abspath(k))
+                    else:
+                        print("Unsupport tag: %s" % v)
+                        sys.exit(-1)
+                    res["filelist"].append("%s%s\n" % (tag, os.path.abspath(k)))
             else:
-                res.append(os.path.abspath(path) + "\n")
+                res["filelist"].append(os.path.abspath(path) + "\n")
 
+        res["filelist"] = inc_dir + res["filelist"]
+        duplicates = set([x for x in res["filelist"] if res["filelist"].count(x) > 1])
+        if duplicates:
+            print("WARN: %s" % " ".join(duplicates))
         return res
 
     def gen_filelist_define(self, defines):
@@ -176,33 +214,68 @@ class ScratChip:
     def gen_filelist(self, cfg_path, dest_target):
         cfg = self.read_yaml(cfg_path)
         targets = {}
+        yaml_targets = {}
         for target, v in cfg["filelist"].items():
             targets[target] = []
+            yaml_targets[target] = {
+                "dirs"  : [],
+                "files" : [],
+            }
             if "defines" in v:
                 targets[target] += self.gen_filelist_define(v["defines"])
             if "files" in v:
-                targets[target] += self.gen_filelist_str(v["files"])
+                res = self.gen_filelist_str(v["files"])
+                yaml_targets[target]["files"] += res["files"]
+                yaml_targets[target]["dirs" ] += res["dirs" ]
+                targets[target] += res["filelist"]
             if "includes" in v:
                 for inc_path in v["includes"]:
                     inc_cfg = self.read_yaml(inc_path)
                     if "defines" in inc_cfg:
                         targets[target] += self.gen_filelist_define(inc_cfg["defines"])
                     if "files" in inc_cfg:
-                        targets[target] += self.gen_filelist_str(inc_cfg["files"])
+                        res =  self.gen_filelist_str(inc_cfg["files"])
+                        yaml_targets[target]["files"] += res["files"]
+                        yaml_targets[target]["dirs" ] += res["dirs" ]
+                        targets[target] += res["filelist"]
 
+        # Filesets Logic
         for target, v in cfg["filelist"].items():
             if "filesets" in v:
+                filesets = []
+                yaml_filesets = {}
                 for other_target in v["filesets"]:
-                    targets[target] = targets[other_target] + targets[target]
+                    filesets += targets[other_target]
+                    yaml_targets[target] = {**yaml_targets[other_target], **yaml_targets[target]}
+                targets[target] = filesets + targets[target]
+
+        # Remove specified line by keyword
+        def remove_line(key_list, line):
+            for k in key_list:
+              if k in line:
+                  return []
+            return line
+        for target, v in cfg["filelist"].items():
+            if "exclude" in v:
+                result = []
+                for line in targets[target]:
+                    result += remove_line(v["exclude"], line)
+                targets[target] = result
 
         dest_dir = "builds/filelist"
-        if not os.path.exists(dest_dir):
-            os.mkdir(dest_dir)
+
+        #if not os.path.exists(dest_dir):
+            #os.mkdir(dest_dir)
+        pathlib.Path(dest_dir).mkdir(parents=True, exist_ok=True)
         if dest_target == "all":
             for target, v in targets.items():
+                with open("%s/%s.yaml" % (dest_dir, target), 'w') as f:
+                    yaml.dump(yaml_targets[target], f)
                 with open("%s/%s.f" % (dest_dir, target), 'w') as f:
                     f.writelines(v)
         elif dest_target in targets:
+            with open("%s/%s.yaml" % (dest_dir, dest_target), 'w') as f:
+                yaml.dump(yaml_targets[dest_target], f)
             with open("%s/%s.f" % (dest_dir, dest_target), 'w') as f:
                 f.writelines(targets[dest_target])
         else:
@@ -261,12 +334,13 @@ def parse_args():
 
     parser_dump_cfg.add_argument(
         'dump_name', type=str, nargs='?',
-        default='config', help='Dump Configure file name ')
+        default='config.yml', help='Dump Configure file name ')
     parser_dump_cfg.set_defaults(func=dump_cfg)
 
     # generate filelist
-    parser_filelist = subparsers.add_parser(
-        "filelist", help="Generate Project's filelist"
+    parser_filelist = subparsers.add_parser("filelist", help="""
+        is_library_file: -v, is_library_dir: -y is_include_dir: +incdir+
+    """
     )
     parser_filelist.add_argument(
         'project_cfg', type=str, nargs='?',
@@ -308,8 +382,7 @@ def init(args):
 def dump_cfg(args):
     cfg = get_resource_name("assets/default.yaml")
     dump_name = args.dump_name
-    sc = ScratChip('.', cfg, dump_name)
-    sc.dump_default_cfg()
+    shutil.copyfile(cfg, dump_name)
 
 def gen_filelist(args):
     cfg = get_resource_name("assets/default.yaml")
